/* 
 * implementation of an Shadow Copy module for use with ZFS snapshots
 *
 * Copyright (C) Stefan Metzmacher	2003-2004
 *
 * Modifications for use with ZFS added by Ed Plese <ed@edplese.com>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "includes.h"

/*
 * This module is based off of the original shadow_copy module with
 * changes to search for the snapshot directories in a subdirectory of the
 * share rather than the root of the share.  This allows ZFS snapshots
 * to be used as shadow copies without the need to create symbolic links
 * to them.
 *
 * ZFS snapshots must be created with names similar in form to the original.
 *
 * Creating a snapshot is as simple as:
 *
 *   # zfs snapshot pool/fs@GMT-`date -u +%Y.%m.%d-%H.%M.%S`
 *
 * Note: Files must differ to be displayed via Windows Explorer!
 * Directories are always displayed...    
*/

static int vfs_shadow_copy_zfs_debug_level = DBGC_VFS;

#undef DBGC_CLASS
#define DBGC_CLASS vfs_shadow_copy_zfs_debug_level

#define SHADOW_COPY_ZFS_PREFIX "@GMT-"

static int shadow_copy_zfs_label_cmp_asc(const void *x, const void *y)
{
	return strncmp((char *)x, (char *)y, sizeof(SHADOW_COPY_LABEL));
}

static int shadow_copy_zfs_label_cmp_desc(const void *x, const void *y)
{
	return -strncmp((char *)x, (char *)y, sizeof(SHADOW_COPY_LABEL));
}

static void shadow_copy_zfs_sort_data(vfs_handle_struct *handle, SHADOW_COPY_DATA *shadow_copy_data)
{
	const char *tmp_str = NULL;

	if (shadow_copy_data && shadow_copy_data->num_volumes > 0 &&
			shadow_copy_data->labels) {

		tmp_str = lp_parm_const_string(SNUM(handle->conn),
			"shadow_copy_zfs", "sort", "");

		if (strcmp(tmp_str, "asc") == 0) {
			qsort(shadow_copy_data->labels,
				shadow_copy_data->num_volumes,
				sizeof(SHADOW_COPY_LABEL),
				shadow_copy_zfs_label_cmp_asc);

		} else if (strcmp(tmp_str, "desc") == 0) {
			qsort(shadow_copy_data->labels,
				shadow_copy_data->num_volumes,
				sizeof(SHADOW_COPY_LABEL),
				shadow_copy_zfs_label_cmp_desc);
		}
	}
 
	return;
}

static BOOL shadow_copy_zfs_match_snapshot_name(const char *name)
{
	/* ignore the leading @ in SHADOW_COPY_ZFS_PREFIX */
	if (strncmp(SHADOW_COPY_ZFS_PREFIX+1, name, sizeof(SHADOW_COPY_ZFS_PREFIX)-2) == 0) {
		return True;
	}

	return False;
}

static BOOL shadow_copy_zfs_match_file_name(const char *name)
{
	if (strncmp(SHADOW_COPY_ZFS_PREFIX, name, sizeof(SHADOW_COPY_ZFS_PREFIX)-1) == 0) {
		return True;
	}

	return False;
}

static SMB_STRUCT_DIR *shadow_copy_zfs_opendir(vfs_handle_struct *handle, connection_struct *conn, const char *fname, const char *mask, uint32 attr)
{
	char modpath[FILENAME_MAX];

	if (shadow_copy_zfs_match_file_name(fname)) {
		snprintf(modpath, FILENAME_MAX, ".zfs/snapshot/%s", &fname[1]);
	} else {
		snprintf(modpath, FILENAME_MAX, "%s", fname);
	}

	DEBUG(10, ("shadow_copy_zfs_opendir: [%s] -> [%s]\n", fname, modpath));

	SMB_STRUCT_DIR *p = SMB_VFS_NEXT_OPENDIR(handle,conn,modpath,mask,attr);

	if (!p) {
		DEBUG(2, ("shadow_copy_zfs_opendir: SMB_VFS_NEXT_OPENDIR() failed for [%s]\n", modpath));
		return NULL;
	}

	return p;
}

static int shadow_copy_zfs_get_shadow_copy_data(vfs_handle_struct *handle, files_struct *fsp, SHADOW_COPY_DATA *shadow_copy_data, BOOL labels)
{
	char modpath[FILENAME_MAX];

	snprintf(modpath, FILENAME_MAX, "%s/.zfs/snapshot", fsp->conn->connectpath);
	DEBUG(10, ("shadow_copy_zfs_get_shadow_copy_data: [%s] -> [%s]\n",
		fsp->conn->connectpath, modpath));
	SMB_STRUCT_DIR *p = SMB_VFS_NEXT_OPENDIR(handle, fsp->conn, modpath,
		NULL, 0);

	shadow_copy_data->num_volumes = 0;
	shadow_copy_data->labels = NULL;

	if (!p) {
		DEBUG(2, ("shadow_copy_zfs_get_shadow_copy_data: SMB_VFS_NEXT_OPENDIR() failed for [%s]\n", modpath));
		return -1;
	}

	while (True) {
		SHADOW_COPY_LABEL *tlabels;
		SMB_STRUCT_DIRENT *d;

		d = SMB_VFS_NEXT_READDIR(handle, fsp->conn, p);
		if (d == NULL) {
			break;
		}

		if (!shadow_copy_zfs_match_snapshot_name(d->d_name)) {
			DEBUG(10, ("shadow_copy_zfs_get_shadow_copy_data: ignore [%s]\n", d->d_name));
			continue;
		}

		DEBUG(10, ("shadow_copy_zfs_get_shadow_copy_data: not ignore [%s]\n", d->d_name));

		if (!labels) {
			shadow_copy_data->num_volumes++;
			continue;
		}

		tlabels = (SHADOW_COPY_LABEL *)TALLOC_REALLOC(shadow_copy_data->mem_ctx,
									shadow_copy_data->labels,
									(shadow_copy_data->num_volumes+1)*sizeof(SHADOW_COPY_LABEL));
		if (tlabels == NULL) {
			DEBUG(2, ("shadow_copy_zfs_get_shadow_copy_data: Out of memory\n"));
			SMB_VFS_NEXT_CLOSEDIR(handle, fsp->conn, p);
			return -1;
		}

		snprintf(tlabels[shadow_copy_data->num_volumes++], sizeof(*tlabels), "@%s", d->d_name);

		shadow_copy_data->labels = tlabels;
	}

	shadow_copy_zfs_sort_data(handle, shadow_copy_data);

	SMB_VFS_NEXT_CLOSEDIR(handle, fsp->conn, p);
	return 0;
}

static int shadow_copy_zfs_stat(vfs_handle_struct *handle, connection_struct *conn, const char *fname, SMB_STRUCT_STAT *sbuf)
{
	char modpath[FILENAME_MAX];

	if (shadow_copy_zfs_match_file_name(fname)) {
		snprintf(modpath, FILENAME_MAX, ".zfs/snapshot/%s", &fname[1]);
	} else {
		snprintf(modpath, FILENAME_MAX, "%s", fname);
	}

	DEBUG(10, ("shadow_copy_zfs_stat: [%s] -> [%s]\n", fname, modpath));

	return SMB_VFS_NEXT_STAT(handle, conn, modpath, sbuf);
}

static int shadow_copy_zfs_open(vfs_handle_struct *handle, connection_struct *conn, const char *fname, int flags, mode_t mode)
{
	char modpath[FILENAME_MAX];

	if (shadow_copy_zfs_match_file_name(fname)) {
		snprintf(modpath, FILENAME_MAX, ".zfs/snapshot/%s", &fname[1]);
	} else {
		snprintf(modpath, FILENAME_MAX, "%s", fname);
	}

	DEBUG(10, ("shadow_copy_zfs_open: [%s] -> [%s]\n", fname, modpath));

	return SMB_VFS_NEXT_OPEN(handle, conn, modpath, flags, mode);
}

/* VFS operations structure */

static vfs_op_tuple shadow_copy_zfs_ops[] = {
	{SMB_VFS_OP(shadow_copy_zfs_opendir),
		SMB_VFS_OP_OPENDIR,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(shadow_copy_zfs_stat),
		SMB_VFS_OP_STAT,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(shadow_copy_zfs_open),
		SMB_VFS_OP_OPEN,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(shadow_copy_zfs_get_shadow_copy_data),
		SMB_VFS_OP_GET_SHADOW_COPY_DATA,SMB_VFS_LAYER_OPAQUE},
	{SMB_VFS_OP(NULL),
		SMB_VFS_OP_NOOP,		SMB_VFS_LAYER_NOOP}
};

NTSTATUS vfs_shadow_copy_zfs_init(void)
{
	NTSTATUS ret = smb_register_vfs(SMB_VFS_INTERFACE_VERSION,
		"shadow_copy_zfs", shadow_copy_zfs_ops);

	if (!NT_STATUS_IS_OK(ret))
		return ret;

	vfs_shadow_copy_zfs_debug_level = debug_add_class("shadow_copy_zfs");
	if (vfs_shadow_copy_zfs_debug_level == -1) {
		vfs_shadow_copy_zfs_debug_level = DBGC_VFS;
		DEBUG(2, ("%s: Couldn't register custom debugging class!\n",
			"vfs_shadow_copy_zfs_init"));
	} else {
		DEBUG(10, ("%s: Debug class number of '%s': %d\n", 
			"vfs_shadow_copy_zfs_init", "shadow_copy_zfs",
			vfs_shadow_copy_zfs_debug_level));
	}

	return ret;
}
